#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "objz80.h"

extern int pc;
extern int line;

int prefix = 0x00;

char * skipwhite(char * str)
{
    if(isblank(*str)) for(;isblank(*str);str++);
    return(str);
}

void generate(unsigned char icode, struct arg * arg1, struct arg * arg2, int ia1, int ia2)
{

    if(arg2)
	if(arg1->prefixed && arg2->prefixed && (arg1->prefixed != arg2->prefixed))
	{
	    fprintf(stderr,"generate(line %d): Unable to assemble two shifts \n",line);
	    exit(1);
	}

    if(prefix != 0xcb)
    {
	if(arg1->prefixed) putbyte(arg1->prefixed);
	if(arg2)
	{
	    if(arg2->prefixed) if(arg2->prefixed!=arg1->prefixed) putbyte(arg2->prefixed);
	    if(prefix) putbyte(prefix);
	    putbyte(icode | (arg1->value << ia1) | (arg2->value << ia2));
	}
	else
	{
	    if(prefix) putbyte(prefix);
	    putbyte(icode | (arg1->value << ia1));
	}
    
	if(arg1->prefixed)
	{
    	    if(arg1->offset<256) putbyte(arg1->offset);
	}
	else
	    if(arg2) if(arg2->prefixed) if(arg2->offset<256) putbyte(arg2->offset);
    }
    else
    {
	if(arg1->prefixed) putbyte(arg1->prefixed);
	if(arg2) if(arg2->prefixed) if(arg2->prefixed!=arg1->prefixed) putbyte(arg2->prefixed);

	if(prefix) putbyte(prefix);
	
	if(arg1->prefixed)
	{
    	    if(arg1->offset<256) putbyte(arg1->offset);
	}
	else
	    if(arg2) if(arg2->prefixed) if(arg2->offset<256) putbyte(arg2->offset);
	

	if(arg2)
	    putbyte(icode | (arg1->value << ia1) | (arg2->value << ia2));
	else
	    putbyte(icode | (arg1->value << ia1));

    }

    if((arg1->type & AT_MASK) == AT_NUMCONST) putbyte(arg1->value);
    if((arg1->type & AT_MASK) == AT_16BITCONST || (arg1->type & AT_MASK) == AT_MEMPTR)
    {
	if(arg1->type & AT_SYMBOLIC)
	    add_sym_reference(arg1->str, pc);
	else
    	    if(arg1->relocate) add_reloc_entry(pc);
	putword(arg1->value);
    }

    if(arg2)
    {
        if((arg2->type & AT_MASK) == AT_NUMCONST) putbyte(arg2->value);
	if((arg2->type & AT_MASK) == AT_16BITCONST || (arg2->type & AT_MASK) == AT_MEMPTR)
	{
	    if(arg2->type & AT_SYMBOLIC)
		add_sym_reference(arg2->str, pc);
	    else
    		if(arg2->relocate) add_reloc_entry(pc);
	    putword(arg2->value);
	}
    }
}

void assemble_word(char * args)
{
    unsigned char * curr = args;
    unsigned char * next;
    unsigned short k;
    int t,i;
    
    for(t = 0; t == 0;)
    {
	/* eliminate blanks */
	next = index(curr, ',');
	if(next) 
	    *next = 0;
	else
	    t = 1;
	if(isblank(*curr)) for(; isblank(*curr); ) curr++;

	if(isdigit(*curr))
	{
	    k = (strtol(curr, NULL, 0) & 0xffff);
	    putword(k);
	}
	else if(*curr=='%')
	{
	    k = (b2d(curr+1) & 0xffff);
	    putword(k);
	}
	else
	{
	    add_sym_reference(curr, pc);
	    putword(0);
	}
	curr = next+1;
    }
}

void assemble_byte(char * args)
{
    unsigned char * curr = args;
    unsigned char * next;
    unsigned char k;
    int t,i;
    
    for(t = 0; t == 0;)
    {
	/* eliminate blanks */
	if(isblank(*curr)) for(; isblank(*curr); ) curr++;
	if(*curr == '"')
	{
	    for(curr++;*curr;curr++)
	    {
		if(*curr == '"') break;
		putbyte(*curr);
	    }
	    if(!(*curr)) return;
	    next = index(curr, ',');
	    if(next) 
	        *next = 0;
	    else
		t = 1;
	}
	else
	{
	    if(*curr == '%')
		k = b2d(curr+1) & 0xff;
	    else
		k = (strtol(curr, NULL, 0) & 0xff);
	    putbyte(k);
	    next = index(curr, ',');
	    if(next) 
	        *next = 0;
	    else
		t = 1;
	}
	curr = next+1;
    }
}

void assemble_dsbyte(char * args)
{
    args = skipwhite(args);
    add_relsym_reference(args, pc);
    putbyte(0x00);
}


void assemble_ld(char *args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    int atsrc, atdest;
    struct arg asrc, adest;
    
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_ld(line %d): LD requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src += 1; src=skipwhite(src);
    
    if(strcmp(src, "a") == 0 && strcmp(dest, "(bc)") == 0) { putbyte(0x02); return; }
    if(strcmp(src, "a") == 0 && strcmp(dest, "(de)") == 0) { putbyte(0x12); return; }
    if(strcmp(src, "(bc)") == 0 && strcmp(dest, "a") == 0) { putbyte(0x0a); return; }
    if(strcmp(src, "(de)") == 0 && strcmp(dest, "a") == 0) { putbyte(0x1a); return; }
    
    if(strcmp(src,"r") ==0 && strcmp(dest,"a") == 0) { putbyte(0xed); putbyte(0x5f); return;}
    if(strcmp(src,"i") ==0 && strcmp(dest,"a") == 0) { putbyte(0xed); putbyte(0x57); return;}
    if(strcmp(src,"a") ==0 && strcmp(dest,"r") == 0) { putbyte(0xed); putbyte(0x4f); return;}
    if(strcmp(src,"a") ==0 && strcmp(dest,"i") == 0) { putbyte(0xed); putbyte(0x47); return;}

    if(strcmp(src,"hl") == 0 && strcmp(dest,"sp") ==0) { putbyte(0xf9); return; }
    if(strcmp(src,"ix") == 0 && strcmp(dest,"sp") ==0) { putbyte(0xdd); putbyte(0xf9); return; }
    if(strcmp(src,"iy") == 0 && strcmp(dest,"sp") ==0) { putbyte(0xfd); putbyte(0xf9); return; }


    parsearg(src, &asrc); parsearg(dest, &adest);

    if((adest.type & AT_MASK) == AT_MEMPTR && strcmp(src,"hl")==0) { generate(0x22,&adest,NULL,8,0); return; }
    if((asrc.type & AT_MASK) == AT_MEMPTR && strcmp(dest,"hl")==0) { generate(0x2a,&asrc,NULL,8,0); return; }
    if((adest.type & AT_MASK) == AT_MEMPTR && strcmp(src,"ix")==0) { putbyte(0xdd); generate(0x22,&adest,NULL,8,0); return; }
    if((asrc.type & AT_MASK) == AT_MEMPTR && strcmp(dest,"ix")==0) { putbyte(0xdd); generate(0x2a,&asrc,NULL,8,0); return; }
    if((adest.type & AT_MASK)  == AT_MEMPTR && strcmp(src,"iy")==0) { putbyte(0xfd); generate(0x22,&adest,NULL,8,0); return; }
    if((asrc.type & AT_MASK) == AT_MEMPTR && strcmp(dest,"iy")==0) { putbyte(0xfd); generate(0x2a,&asrc,NULL,8,0); return; }

    
    if(asrc.type == AT_8BITREG && adest.type == AT_8BITREG) { generate(0x40,&asrc,&adest,0,3); return; }
    if(asrc.type == AT_NUMCONST && adest.type == AT_8BITREG) { generate(0x06,&asrc,&adest,8,3); return; }

    
    if(adest.type == AT_16BITREG)
    {
	if(asrc.type == AT_NUMCONST) asrc.type = AT_16BITCONST;
	if((asrc.type & AT_MASK) == AT_16BITCONST) { generate(0x01,&asrc,&adest,8,4); return; }
	if((asrc.type & AT_MASK) == AT_MEMPTR) { prefix=0xed; generate(0x4b,&asrc,&adest,8,4); return; }
    }
    if((adest.type & AT_MASK) == AT_MEMPTR && asrc.type == AT_16BITREG) { prefix=0xed; generate(0x43,&asrc,&adest,4,8); return; }
    
    if((asrc.type & AT_MASK) == AT_MEMPTR && strcmp(dest,"a") == 0) { generate(0x3a,&asrc,NULL,8,0); return; }
    if((adest.type & AT_MASK) == AT_MEMPTR && strcmp(src,"a") == 0) { generate(0x32,&adest,NULL,8,0); return; }
    fprintf(stderr,"assemble_ld(line %d): Unsupported set of operands\n",line);
    exit(1);
}

void assemble_push(char *args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type != AT_16BITREG) { fprintf(stderr,"assemble_push(line %d): Invalid operand type\n",line); exit(0); }
    generate(0xc5, &arg, NULL, 4, 0);
}

void assemble_pop(char *args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type != AT_16BITREG) { fprintf(stderr,"assemble_pop(line %d): Invalid operand type\n",line); exit(0); }
    generate(0xc1, &arg, NULL, 4, 0);
}

void assemble_exx(char * args)
{
    putbyte(0x09);
}

void assemble_ex(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_ex(line %d): EX requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);
    if(strcmp(src,"af'") == 0 && strcmp(dest,"af") ==0) { putbyte(0x08); return; }
    if(strcmp(src,"hl") == 0 && strcmp(dest,"de") ==0) { putbyte(0xeb); return; }
    if(strcmp(src,"hl") == 0 && strcmp(dest,"(sp)") ==0) { putbyte(0xe3); return; }
    if(strcmp(src,"ix") == 0 && strcmp(dest,"(sp)") ==0) { putbyte(0xdd); putbyte(0xe3); return; }
    if(strcmp(src,"iy") == 0 && strcmp(dest,"(sp)") ==0) { putbyte(0xfd); putbyte(0xe3); return; }
    fprintf(stderr,"assemble_ex(line %d): Invalid operand set\n",line);
    exit(0);
}

void assemble_add(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_add(line %d): ADD requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc); parsearg(dest, &adest);
    if(strcmp(dest,"a")==0)
    {
	if(asrc.type == AT_8BITREG)
	{
	    generate(0x80,&asrc,NULL,0,0); return;
	}
	if(asrc.type == AT_NUMCONST)
	{
	    generate(0xc6,&asrc,NULL,8,0); return;
	}
    }
    if(adest.value = 6 && adest.type == AT_16BITREG && asrc.type == AT_16BITREG) 
	{ generate(0x09,&asrc,&adest,4,8); return; }
    fprintf(stderr,"assemble_add(line %d): Invalid set of operands\n",line); exit(1);
}

void assemble_adc(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_adc(line %d): ADC requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc); parsearg(dest, &adest);
    if(strcmp(dest,"a")==0)
    {
	if(asrc.type == AT_8BITREG)
	{
	    generate(0x88,&asrc,NULL,0,0); return;
	}
	if(asrc.type == AT_NUMCONST)
	{
	    generate(0xce,&asrc,NULL,8,0); return;
	}
    }
    if(strcmp(dest,"hl")==0 && asrc.type==AT_16BITREG)
	{ prefix=0xed; generate(0x4a,&asrc,NULL,4,0); return; }
    fprintf(stderr,"assemble_adc(line %d): Invalid set of operands\n",line); exit(1);
}

void assemble_sbc(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_sbc(line %d): SBC requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc); parsearg(dest, &adest);
    if(strcmp(dest,"a")==0)
    {
	if(asrc.type == AT_8BITREG)
	{
	    generate(0x98,&asrc,NULL,0,0); return;
	}
	if(asrc.type == AT_NUMCONST)
	{
	    generate(0xde,&asrc,NULL,8,0); return;
	}
    }    
    if(strcmp(dest,"hl")==0 && asrc.type==AT_16BITREG)
	{ prefix=0xed; generate(0x42,&asrc,NULL,4,0); return; }
    fprintf(stderr,"assemble_sbc(line %d): Invalid set of operands\n",line); exit(1);
}

void assemble_sub(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_NUMCONST) { generate(0xd6, &arg, NULL, 8, 0); return; }
    if(arg.type == AT_8BITREG)  { generate(0x90, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_sub(line %d): Invalid operand\n",line); exit(1);
}

void assemble_cp(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_NUMCONST) { generate(0xfe, &arg, NULL, 8, 0); return; }
    if(arg.type == AT_8BITREG)  { generate(0xb8, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_cp(line %d): Invalid operand\n",line); exit(1);
}

void assemble_and(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_NUMCONST) { generate(0xe6, &arg, NULL, 8, 0); return; }
    if(arg.type == AT_8BITREG) { generate(0xa0, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_and(line %d): Invalid operand\n",line); exit(1);
}


void assemble_or(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_NUMCONST) { generate(0xf6, &arg, NULL, 8, 0); return; }
    if(arg.type == AT_8BITREG) { generate(0xb0, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_or(line %d): Invalid operand\n",line); exit(1);
}

void assemble_xor(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_NUMCONST) { generate(0xee, &arg, NULL, 8, 0); return; }
    if(arg.type == AT_8BITREG) { generate(0xa8, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_xor(line %d): Invalid operand\n",line); exit(1);
}

void assemble_inc(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { generate(0x04, &arg, NULL, 3, 0); return; }
    if(arg.type == AT_16BITREG) { generate(0x03, &arg, NULL, 4, 0); return; }
    fprintf(stderr,"assemble_inc(line %d): Invalid operand\n",line); exit(1);
}

void assemble_dec(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { generate(0x05, &arg, NULL, 3, 0); return; }
    if(arg.type == AT_16BITREG) { generate(0x0b, &arg, NULL, 4, 0); return; }
    fprintf(stderr,"assemble_dec(line %d): Invalid operand\n",line); exit(1);
}


void assemble_neg(char *args)
{
    putbyte(0xed);
    putbyte(0x44);
}
void assemble_cpl(char *args) { putbyte(0x2f); }
void assemble_scf(char *args) { putbyte(0x37); }
void assemble_ccf(char *args) { putbyte(0x3f); }
void assemble_daa(char *args) { putbyte(0x27); }

void assemble_rlca(char * args) { putbyte(0x07); }
void assemble_rrca(char * args) { putbyte(0x0f); }
void assemble_rla(char * args) { putbyte(0x17); }
void assemble_rra(char * args) { putbyte(0x1f); }

void assemble_rlc(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x00, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_rlc(line %d): Invalid operand\n",line); exit(1);
}

void assemble_rrc(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x08, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_rrc(line %d): Invalid operand\n",line); exit(1);
}

void assemble_rl(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x10, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_rl(line %d): Invalid operand\n",line); exit(1);
}

void assemble_rr(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x18, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_rr(line %d): Invalid operand\n",line); exit(1);
}

void assemble_sla(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x20, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_sla(line %d): Invalid operand\n",line); exit(1);
}

void assemble_sra(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x28, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_sra(line %d): Invalid operand\n",line); exit(1);
}

void assemble_sli(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x30, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_sli(line %d): Invalid operand\n",line); exit(1);
}

void assemble_srl(char * args)
{
    struct arg arg;
    parsearg(skipwhite(args), &arg);
    if(arg.type == AT_8BITREG) { prefix=0xcb; generate(0x38, &arg, NULL, 0, 0); return; }
    fprintf(stderr,"assemble_srl(line %d): Invalid operand\n",line); exit(1);
}

void assemble_rrd(char * args) { putbyte(0xed); putbyte(0x67); }
void assemble_rld(char * args) { putbyte(0xed); putbyte(0x6f); }

void assemble_ldi(char * args) { putbyte(0xed); putbyte(0xa0); }
void assemble_ldd(char * args) { putbyte(0xed); putbyte(0xab); }
void assemble_ldir(char * args) { putbyte(0xed); putbyte(0xb0); }
void assemble_lddr(char * args) { putbyte(0xed); putbyte(0xb8); }

void assemble_cpir(char * args) { putbyte(0xed); putbyte(0xb1); }
void assemble_cpdr(char * args) { putbyte(0xed); putbyte(0xb9); }

void assemble_nop(char * args) { putbyte(0x00); }
void assemble_halt(char * args) { putbyte(0x76); }
void assemble_di(char * args) { putbyte(0xf3); }
void assemble_ei(char * args) { putbyte(0xfb); }


void assemble_im(char * args)
{
    putbyte(0xed);
    switch(atoi(args))
    {
	case 0: putbyte(0x46); return;
	case 1: putbyte(0x56); return;
	case 2: putbyte(0x5e); return;
    }
    fprintf(stderr,"assemble_im(line %d): Incorrect interrupt handling mode\n",line);
}

int decode_jump(char * code)
{
    if(strcmp(code, "nz")==0) return(0);
    if(strcmp(code, "z")==0) return(1);
    if(strcmp(code, "nc")==0) return(2);
    if(strcmp(code, "c")==0) return(3);
    if(strcmp(code, "po")==0) return(4);
    if(strcmp(code, "pe")==0) return(5);
    if(strcmp(code, "p")==0) return(6);
    if(strcmp(code, "m")==0) return(7);
    return(-1);
}

void assemble_jp(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    int jc;
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	parsearg(dest, &adest);
	if((adest.type & AT_MASK) == AT_MEMPTR)
	{
	    generate(0xc3, &adest, NULL, 8, 0);
	    return;
	}
	else if(strcmp(dest,"(hl)")==0)
	{
	    putbyte(0xe9);
	    return;
	}
	else if(strcmp(dest,"(ix)")==0)
	{
	    putbyte(0xdd);
	    putbyte(0xe9);
	    return;
	}
	else if(strcmp(dest,"(iy)")==0)
	{
	    putbyte(0xfd);
	    putbyte(0xe9);
	    return;
	}
	fprintf(stderr,"assemble_jp(line %d): Invalid set of operand[s]\n",line); exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc); parsearg(dest, &adest);
    
    if((asrc.type & AT_MASK)  == AT_MEMPTR)
    {
	jc=decode_jump(dest);
	if(jc == -1) { fprintf(stderr,"assemble_jp(line %d): Invalid jump condition\n",line); exit(1); }
	generate(0xc2 | (jc << 3), &asrc, NULL , 8, 0);
	return;
    }
    
    fprintf(stderr,"assemble_jp(line %d): Invalid set of operand[s]\n",line); exit(1);
}

void assemble_jr(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    int jc;
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	putbyte(0x18);
	add_relsym_reference(dest, pc);
	putbyte(00);
	return;
    }
    *src = 0;
    src ++;
    src = skipwhite(src);
    if(strcmp(dest,"nz")==0)
    {
	putbyte(0x20); add_relsym_reference(src, pc); putbyte(00); return;
    }
    if(strcmp(dest,"z")==0)
    {
	putbyte(0x28); add_relsym_reference(src, pc); putbyte(00); return;
    }
    if(strcmp(dest,"nc")==0)
    {
	putbyte(0x30); add_relsym_reference(src, pc); putbyte(00); return;
    }
    if(strcmp(dest,"c")==0)
    {
	putbyte(0x38); add_relsym_reference(src, pc); putbyte(00); return;
    }

    fprintf(stderr,"assemble_jr(line %d): Invalid set of operand[s] (dest '%s')\n",line, dest); exit(1);
}

void assemble_djnz(char * args)
{
    unsigned char * dest = skipwhite(args);
    putbyte(0x10); add_relsym_reference(dest, pc); putbyte(00);
}

void assemble_call(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    int jc;
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	parsearg(dest, &adest);
	if((adest.type & AT_MASK) == AT_MEMPTR)
	{
	    generate(0xcd, &adest, NULL, 8, 0);
	    return;
	}
	fprintf(stderr,"assemble_call(line %d): Invalid set of operand[s]\n",line); exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc); parsearg(dest, &adest);
    
    if((asrc.type & AT_MASK) == AT_MEMPTR)
    {
	jc=decode_jump(dest);
	if(jc == -1) { fprintf(stderr,"assemble_call(line %d): Invalid jump condition\n",line); exit(1); }
	generate(0xc4 | (jc << 3), &asrc, NULL , 8, 0);
	return;
    }
    
    fprintf(stderr,"assemble_call(line %d): Invalid set of operand[s]\n",line); exit(1);
}

void assemble_rst(char *args)
{
    unsigned char * dest = skipwhite(args);
    int q, k;
    q = strtol(dest, NULL, 0);
    if(((q/8)*8) != q || ((k=q/8) > 7))
    {
	fprintf(stderr,"assemble_rst(line %d): Invalid restart code\n",line); exit(0);
    }
    putbyte(0xc7 | (k << 3));
}

void assemble_ret(char * args)
{
    unsigned char * dest = skipwhite(args);
    int jc;
    
    
    if(strlen(dest) == 0)
	putbyte(0xc9);
    else
    {
	jc=decode_jump(dest);
	if(jc == -1) { fprintf(stderr,"assemble_ret(line %d): Invalid jump condition\n",line); exit(1); }
	putbyte(0xc0 | (jc << 3));
    }
}

void assemble_reti(char * args) { putbyte(0xed); putbyte(0x4d); }
void assemble_retn(char * args) { putbyte(0xed); putbyte(0x45); }

void assemble_bit(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    int b;

    prefix = 0xcb;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_bit(line %d): BIT requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);
    
    b=atoi(dest);
    if(b>7 || b<0)
    {
	fprintf(stderr,"assemble_bit(line %d): Z80 registers are 8-bit\n",line);
	exit(1);
    }
    parsearg(src, &asrc);
    if(asrc.type != AT_8BITREG) 
    {
	fprintf(stderr,"assemble_bit(line %d): Invalid operand set\n",line);
	exit(1);
    }
    generate(0x40 | (b <<3), &asrc, NULL, 0, 0);
}


void assemble_set(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    int b;

    prefix = 0xcb;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_set(line %d): SET requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);
    
    b=atoi(dest);
    if(b>7 || b<0)
    {
	fprintf(stderr,"assemble_set(line %d): Z80 registers are 8-bit\n",line);
	exit(1);
    }
    parsearg(src, &asrc);
    if(asrc.type != AT_8BITREG) 
    {
	fprintf(stderr,"assemble_set(line %d): Invalid operand set\n",line);
	exit(1);
    }
    generate(0xc0 | (b <<3), &asrc, NULL, 0, 0);
}


void assemble_res(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    int b;
    
    prefix = 0xcb;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_res(line %d): RES requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);
    
    b=atoi(dest);
    if(b>7 || b<0)
    {
	fprintf(stderr,"assemble_res(line %d): Z80 registers are 8-bit\n",line);
	exit(1);
    }
    parsearg(src, &asrc);
    if(asrc.type != AT_8BITREG) 
    {
	fprintf(stderr,"assemble_res(line %d): Invalid operand set\n",line);
	exit(1);
    }
    generate(0x80 | (b <<3), &asrc, NULL, 0, 0);
}

void assemble_ini(char * args) { putbyte(0xed); putbyte(0xa2); }
void assemble_ind(char * args) { putbyte(0xed); putbyte(0xaa); }
void assemble_outi(char * args) { putbyte(0xed); putbyte(0xa3); }
void assemble_outd(char * args) { putbyte(0xed); putbyte(0xab); }
void assemble_inir(char * args) { putbyte(0xed); putbyte(0xb2); }
void assemble_indr(char * args) { putbyte(0xed); putbyte(0xba); }
void assemble_otir(char * args) { putbyte(0xed); putbyte(0xb3); }
void assemble_otdr(char * args) { putbyte(0xed); putbyte(0xbb); }

void assemble_in(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_in(line %d): IN requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(dest, &adest);


    if(strcmp(src,"(c)")==0)
    {
	if(adest.type == AT_8BITREG)
	{
	    putbyte(0xed);
	    putbyte(0x40 | (adest.value << 3));
	    return;
	}
    }
    else
    {
        parsearg(src, &asrc);
	if(asrc.type == AT_MEMPTR && strcmp(dest,"a")==0)
	{
	    putbyte(0xdb);
	    putbyte(asrc.value);
	    return;
	}
    }
    
    fprintf(stderr,"assemble_in(line %d): Invalid set of operands\n",line); exit(1);
}

void assemble_out(char * args)
{
    unsigned char * src;
    unsigned char * dest = skipwhite(args);
    struct arg asrc;
    struct arg adest;
    
    src = index(args,',');
    if(!src)
    {
	fprintf(stderr,"assemble_out(line %d): OUT requires two operands\n",line);
	exit(1);
    }
    *src = 0;
    src=skipwhite(src+1);

    parsearg(src, &asrc);

    if(strcmp(dest,"(c)")==0)
    {
	if(asrc.type == AT_8BITREG)
	{
	    putbyte(0xed);
	    putbyte(0x41 | (adest.value << 3));
	    return;
	}
    }
    else
    {
	parsearg(dest, &adest);
	if(adest.type == AT_MEMPTR && strcmp(src,"a")==0)
	{
	    putbyte(0xd3);
	    putbyte(adest.value);
	    return;
	}
    }
    
    fprintf(stderr,"assemble_out(line %d): Invalid set of operands\n",line); exit(1);
}
