
# switch.asm/esOS

# s_interrupt		PERFORM TASK SWITCHING
#
#
#

s_interrupt
; PERFORM MULTITASKING - TASK SWITCHING
; 1) save state of the old task
    push bc
    push de
    push hl
    push ix
    push iy			; save registers
    push af
    
; 2) get stack frame pointer
    ld hl, #0
    add hl, sp			; get SP into HL
    
    push hl
    pop bc			; and copy it into BC
; 3) get current task structure
    ld hl, (#24004)		; current task
; 4) get next task structure pointer
    ld e, (hl)
    inc hl
    ld d, (hl)			; into DE
    inc hl
; 5) save old task stack pointer
    ld (hl), c
    inc hl
    ld (hl), b
; 6) call i/o routines
    call (s_console_int)
; 7) change task
s_task_ignite
    push de
    pop hl

    ld (#24004), hl		; update system parameters
    inc hl
    inc hl			; skip next task pointer
    ld e, (hl)
    inc hl
    ld d, (hl)			; get stack frame pointer
; 8) update stack frame
    ex de, hl
    ld sp, hl			; load new stack
; 9) restore new process state
    pop af
    pop iy
    pop ix
    pop hl
    pop de
    pop bc
; 10) continue execution
    ei
    ret

# s_task_create    CREATE A TASK STRUCTURE
#
#		BC - initial PC
#	returns HL - task structure
#


s_task_create


    ld ix, #0
    add ix, sp		; get the stack pointer
    
    push bc		; save routine parameter
    
    ld de, #256
    call (s_alloc)	; allocate memory for stack & new process structure
    ld de, #250
    add hl, de		; get structure pointer
    
    pop bc		; restore BC
    
    push hl
    pop de		; get structure pointer in DE

    
    ld sp, hl		; load stack
    push bc		; load PC
    push af
    push af
    push af
    push af
    push af
    push af		; load fake registers

    ld hl, #0
    add hl, sp		; stack pointer

    push hl
    pop bc		; now the stack pointer is at BC
    
    push de
    pop hl		; DE holded structure pointer, copy it into HL
    ld (hl), e
    inc hl
    ld (hl), d
    inc hl		; load our pointer as next_task pointer
    ld (hl), c
    inc hl
    ld (hl), b		; load stack pointer
    inc hl
    ld bc, (#24154)     ; current console
    ld (hl), c
    inc hl
    ld (hl), b		; save console pointer

    
    push ix
    pop hl		; load old stack ptr from IX to HL

    ld sp, hl		; restore old stack pointer

    ex de, hl
    ret

# s_task_add	CREATE A TASK STRUCTURE AND RUN IT
#
#		BC	- initial PC
#

s_task_link
    push hl
    pop de			; copy data to DE
    
    di				; we'll be performing some critical operations ...
    ld iy, (#24002)		; get the last task

    ld (iy+0), e
    ld (iy+1), d		; we are now it's next task
    
    ld (#24002), hl		; however, we are the last task now
    push hl
    pop iy			; copy our structure to IY
    
    ld hl, (#24000)		; get the first task
    
    push hl
    pop de
    ld (iy+0), e
    ld (iy+1), d		; it'll be ours next task ...
    ret

s_task_add
    call (s_task_create)	; now HL holds task structure pointer
    call (s_task_link)
    im 1
    ei				; now we are ready to run
    ret


;	we assume that interrupts are disabled

s_task_unlink
    ld hl, (#24004)		; base->current_task
    push hl
    ld a, h
rt_prevtaskloop
    ld e, (hl)
    inc hl			; skip lower part of next_task
    ld d, (hl)			; get in DE next task pointer
    cp d			; next_task == we?
    ex de, hl			; next element in list
    jr nz, rt_prevtaskloop	; next task != we.
    dec de			; now in DE our previous task pointer
    pop hl			; now in HL -> our task
    push de			; store our previous task
    push hl			; store our task
    ldi
    ldi				; our next task is now it's next task
    

    ld hl, (#24002)		; last task
    pop de			; restore our task
    
    ld a, d
    cp h			; we ?
    pop hl			; restore our previous task
    jr nz, rt_done		; not we.
    
    ld (#24002), hl		; our previous task will be the last task then
rt_done
    ret
    
s_task_suspend
    di
    call (s_task_unlink)
    call (s_interrupt)		; perform task switching
    ret				; we've been bringed back to life

s_task_remtask
    di
    call (s_task_unlink)
    ld hl, (#24004)
    ld e, (hl)
    inc hl
    ld d, (hl)

    push de
    dec hl
    call (s_free)
    pop de

    jp (s_task_ignite)


!global s_interrupt
!global s_task_create
!global s_task_add
!global s_task_ignite
!global s_task_unlink
!global s_task_suspend
!global s_task_remtask
!global s_task_link
